@@ -1,6 +1,10 @@
 """
 
 """
+import textwrap
+from pprint import pformat
+from typing import List, Union, Any
+
 import copy
 import json
 import os
@@ -113,15 +117,34 @@ class BaseParams():
         self._param_dict = d['_param_dict']
         self._repeat = d['_repeat']
         self._lock = d['_lock']
-        self._bind = d['_bound']
+        self._constrain = d['_bound']
         self._namespace = d['_namespace']
 
     def __repr__(self):
+        import textwrap
+
         dynamic_propertys = [(k, io.safe_getattr(self, k, None)) for k in self.__dir__() if
                              isinstance(getattr(self.__class__, k, None), property)]
 
-        return "{}".format(self.__class__.__name__) + pp.pformat(
-            [(k, v) for k, v in chain(self._param_dict.items())] + [{'propertys': dynamic_propertys}])
+        dynamic_propertys = [(k, v, f'{type(v).__name__}') for k, v in dynamic_propertys]
+
+        def _arg_to_str(k, v):
+            res = self._constrain.get(k, None)
+            if res is not None:
+                return f'{res}, {type(v).__name__}'
+            return f'{type(v).__name__}'
+
+        args = [(k, v) for k, v in
+                chain(self._param_dict.items())]
+        args = [(k, v, _arg_to_str(k, v)) for k, v in args]
+
+        args_str = BaseParams.safe_param_repr(args)
+
+        if len(dynamic_propertys) > 0:
+            property_str = BaseParams.safe_param_repr(dynamic_propertys)
+            return "{}.Space".format(
+                self.__class__.__name__) + '(\n' + args_str + '\n    # @property\n' + property_str + '\n)'
+        return "{}.Space".format(self.__class__.__name__) + '(\n' + args_str + '\n)'
 
     __str__ = __repr__
 
@@ -146,6 +169,55 @@ class BaseParams():
         elif isinstance(bound, choice_param) and value not in bound.choices:
             raise BoundCheckError(f"value of param '{name}' should in values {bound.choices}.")
 
+    @staticmethod
+    def _safe_repr(values: Any) -> str:
+        return pformat(values)
+
+    @staticmethod
+    def _padding_mod(st: str, offset=7, mod=4):
+        """
+        123 \\
+        1   \\
+        12312341    \\
+        1231
+        Args:
+            strs:
+            mod:
+
+        Returns:
+
+        """
+        size = len(st)
+        if size < offset:
+            return st.ljust(offset, ' ')
+
+        mnum = mod - len(st) % mod
+        # if mnum == 0:
+        #     mnum = mod
+        return st.ljust(size + mnum, ' ')
+
+    @staticmethod
+    def safe_param_repr(values: List[tuple], level=1) -> str:
+        """
+
+        Args:
+            values:
+            level:
+
+        Returns:
+
+        """
+        res = [(f"{k}={BaseParams._safe_repr(v)},", anno) for k, v, anno in values]
+
+        # res = textwrap.fill('\n'.join(res))
+        res = '\n'.join([BaseParams._padding_mod(i, offset=16, mod=4) + f'  # {anno}' for i, anno in res])
+
+        return textwrap.indent(res, '    ')
+
+    @classmethod
+    def Space(cls, **kwargs):
+        return cls().from_dict(kwargs)
+
     def copy(self):
         res = self.__class__()
         res._param_dict = self._param_dict.copy()
@@ -274,6 +346,7 @@ class BaseParams():
         def func(**kwargs):
             if '_help' in kwargs:
                 print(self)
+                exit()
                 return
 
             if '_json' in kwargs:
@@ -291,6 +364,7 @@ class BaseParams():
                 self[k] = v
 
         fire.Fire(func)
+        self.iparams()
         return self
 
     def from_yaml(self, fn):
@@ -316,6 +390,7 @@ class BaseParams():
                 print(e, file=sys.stderr)
         else:
             print(f'{fn} not exists, ignored load params from yaml file, please verify the path.', file=sys.stderr)
+        self.iparams()
         return self
 
     def from_json(self, fn):
@@ -337,6 +412,8 @@ class BaseParams():
                 print(e, file=sys.stderr)
         else:
             print(f'{fn} not exists, ignored load params from json file, please verify the path', file=sys.stderr)
+
+        self.iparams()
         return self
 
     def from_dict(self, dic: dict):
@@ -344,6 +421,9 @@ class BaseParams():
         for k, v in dic.items():
             self[k] = v
 
+        self.iparams()
+        return self
+
     def to_json(self, fn: str):
         """
         write params to the specified file path in json format.
@@ -425,6 +505,9 @@ class BaseParams():
         for param in params:
             self._param_dict.update(param._param_dict)
 
+    def iparams(self):
+        pass
+
 
 class DistributionParams(BaseParams):
     def __init__(self):
@@ -436,7 +519,7 @@ class DistributionParams(BaseParams):
 
     @property
     def init_method(self):
-        from lumo.utils.connect import find_free_network_port
+        from lumo.proc.network import find_free_network_port
         port = find_free_network_port()
         return f'tcp://localhost:{port}'
 
@@ -487,14 +570,7 @@ class Params(BaseParams):
         self.idx = 0
         self.global_step = 0
         self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
-        self.dataset = None
-        self.arch = None
         self.stage = self.choice('init', 'train', 'test', 'val')
-        self.optim = None  # type:OptimBuilder
-        # self.tmp_dir = None  # type:str # set TMPDIR environment
-
-    def iparams(self):
-        pass
 
 
 ParamsType = TypeVar('ParamsType', bound=Params)
