@@ -18,7 +18,7 @@ from torch import distributed as dist
 from torch import nn
 from torch.optim import Optimizer
 from torch.utils.data import DataLoader
-
+from functools import lru_cache
 from .environ import globs
 from .experiment import TrainerExperiment
 from .logger import Logger
@@ -31,7 +31,7 @@ from .mixin import ModelMix, CallbackMix, DataModuleMix
 from .params import ParamsType
 from lumo.base_classes import attr, TrainerStage
 from lumo.base_classes.metaclasses import Merge
-from lumo.utils.keys import TRAINER
+from ..proc.const import TRAINER
 from lumo.utils import dist
 from lumo.utils.device import construct_device_args_kwargs, to_device_enumrate
 from lumo.contrib.accelerate import Accelerator
@@ -48,10 +48,14 @@ class initial_tuple():
 
 
 class TrainerResult(attr):
-    def __init__(self, state: TrainerStage, result: int, msg=''):
+    MSG_OK = 'success'
+    MSG_OK_MID = 'success stop in midway'
+    MSG_NO_DATALOADER = 'dataloader not loaded'
+
+    def __init__(self, state: TrainerStage, meter: AvgMeter, msg=''):
         super().__init__()
         self.state = state
-        self.result = result
+        self.meter = meter
         self.msg = msg
 
 
@@ -180,13 +184,12 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
             'params': params,
             'exp': TrainerExperiment(self._gene_class_exp_name()),
         })
-        params.iparams()
-
         self._initialize_globs()
         self._initial_exp()
 
         self._logger = None
         self._rnd = None
+        self._accelerate = None
         self._saver = None
         self._datamodule = None
 
@@ -205,8 +208,9 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
         super().__setattr__(name, value)
         from torch.optim.optimizer import Optimizer
 
-        if name.startswith('_'):
+        if name.startswith('_') or name.endswith('_'):
             # when callback is trainer itself, _hooked will be passed, and caused recursion
+            # if some pretrained models need to be ignored in save/load stage, it can be named like 'some_' or '_some'
             return
 
         if isinstance(value, torch.device):
@@ -220,7 +224,6 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
         elif isinstance(value, (np.ndarray)):
             self._state_dicts[TRAINER.DKEY.tensor]['np'][name] = value
         elif callable(getattr(value, "state_dict", None)) and callable(getattr(value, "load_state_dict", None)):
-
             self._state_dicts[TRAINER.DKEY.others][name] = value
 
     def __setitem__(self, key: str, value: Any):
@@ -308,7 +311,8 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
             self._logger = Logger()
             set_global_logger(self._logger)
             if self.params.get('debug', False):
-                Logger.set_verbose(Logger.V_DEBUG)
+                self._logger.set_verbose(Logger.V_DEBUG)
+                self._logger.dddebug('Enable debug log.')
             fn = self._logger.add_log_dir(self.exp.log_dir)
             self.exp.writeline('logger', fn)
         return self._logger
@@ -319,17 +323,20 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
 
     @property
     def accelerator(self) -> Accelerator:
-        acce = {'device_placement': True,
-                'split_batches': False,
-                'fp16': None,
-                'cpu': False,
-                'rng_types': None,
-                'kwargs_handlers': None,
-                'device': self.params.get('device', None)}
-        for k in acce:
-            if k in self.params:
-                acce[k] = self.params[k]
-        return Accelerator(**acce)
+        if self._accelerate is None:
+            acce = {'device_placement': True,
+                    'split_batches': False,
+                    'fp16': None,
+                    'cpu': False,
+                    'rng_types': None,
+                    'kwargs_handlers': None,
+                    'device': self.params.get('device', None)}
+            for k in acce:
+                if k in self.params:
+                    acce[k] = self.params[k]
+
+            self._accelerate = Accelerator(**acce)
+        return self._accelerate
 
     @property
     def rnd(self) -> RndManager:
@@ -587,6 +594,9 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
             if k in src:
                 v.load_state_dict(src[k])
 
+    def ignore_object_state_dict(self, item):
+        pass
+
     def load_state_dict(self, object: Union[str, dict]):
         if isinstance(object, str):
             ckpt, meta_info = self.saver.load_state_dict(object, with_meta=True)
@@ -631,7 +641,7 @@ class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
 
     def save_model(self, is_best=False, meta_info: Union[str, dict, Meter] = None):
         info = self._build_trainer_meta_info(meta_info)
-        return self.saver.save_model(self.eidx, self.model_state_dict(),
+        return self.saver.save_model(self.eidx, {'models': self.model_state_dict()},
                                      meta_info=info,
                                      is_best=is_best)
 
@@ -684,7 +694,7 @@ class Trainer(DLLoopMix, _BaseTrainer):
         'prepare_dataloader',
     }
 
-    def _wrap_result(self, meter: Union[Mapping, Meter, Sequence, torch.Tensor, np.ndarray]):
+    def _wrap_result(self, meter: Union[Mapping, Meter, Sequence, torch.Tensor, np.ndarray]) -> dict:
         if meter is None:
             return {}
         if isinstance(meter, (Mapping, Meter)):
@@ -726,28 +736,37 @@ class Trainer(DLLoopMix, _BaseTrainer):
         else:
             return to_device(item, device_args_kwargs)
 
-    def prepare_dataloader(self, stage: TrainerStage, dataloader=None):
+    def prepare_dataloader(self, stage: TrainerStage, dataloader=None) -> DataLoader:
         params = self.params
         initialized = getattr(self.initial, f"{stage.name}_dataloader")
 
+        dataloader_ = None
         """use loader from passed params first, then use itself, finally"""
         if isinstance(dataloader, DataModuleMix):
             dataloader.idataloader(params, stage, initialized)
             self.regist_dataloader(datamodule=dataloader)
-            dataloader = getattr(dataloader, f'{stage.name}_dataloader')
+            dataloader_ = getattr(dataloader, f'{stage.name}_dataloader', None)
+        elif isinstance(dataloader, DataLoader):
+            self.regist_dataloader(**{stage.name: dataloader})
+            dataloader_ = dataloader
         elif isinstance(self, DataModuleMix):
             self.idataloader(params, stage, initialized)
-            dataloader = getattr(self, f'{stage.name}_dataloader')
+            dataloader_ = getattr(self, f'{stage.name}_dataloader', None)
 
-        if dataloader is None:
-            dataloader = getattr(dataloader, f'{stage.name}_dataloader')
-        dataloader = self.accelerator.prepare(dataloader)
+        if dataloader is None and self.datamodule is not None:
+            dataloader_ = getattr(self.datamodule, f'{stage.name}_dataloader', None)
+            if dataloader_ is None:
+                self.datamodule.idataloader(params, stage, initialized)
+                dataloader_ = getattr(self.datamodule, f'{stage.name}_dataloader', None)
 
-        kwargs = {stage.name: dataloader}
+        if isinstance(dataloader_, DataLoader):
+            dataloader_ = self.accelerator.prepare(dataloader_)
+
+        kwargs = {stage.name: dataloader_}
         self.regist_dataloader(**kwargs)
-        return dataloader
+        return dataloader_
 
-    def train(self, dataloader: Union[DataLoader, DataModuleMix] = None):
+    def train(self, dataloader: Union[DataLoader, DataModuleMix] = None) -> TrainerResult:
         # params, initialized = self.params, self.initial.train_dataloader
         dataloader = self.prepare_dataloader(TrainerStage.train, dataloader)
         if dataloader is None:
@@ -756,24 +775,26 @@ class Trainer(DLLoopMix, _BaseTrainer):
         self._check_models_init()
         self._check_optim_init()
 
-        self.to_stage(TrainerStage.train)
-
         self.initial.train_dataloader = True
 
+        result = TrainerResult(TrainerStage.train, None, TrainerResult.MSG_OK)
         while self.params.eidx < self.params.epoch:
             self.params.eidx += 1
-            self.train_epoch(dataloader)
+            self.to_stage(TrainerStage.train)
+            result = self.train_epoch(dataloader)
             if self.train_toggle:
                 self.train_toggle = False
-                return TrainerResult(TrainerStage.train, 1, 'stoped midway')
+                return TrainerResult(TrainerStage.train, result.meter, TrainerResult.MSG_OK_MID)
 
             self.exp.dump_train_info(self.params.eidx)
 
-        return TrainerResult(TrainerStage.train, 0)
+        return TrainerResult(TrainerStage.train, result.meter, TrainerResult.MSG_OK)
 
-    def train_epoch(self, dataloader: DataLoader):
-        avg = AvgMeter()
+    def train_epoch(self, dataloader: DataLoader) -> TrainerResult:
+        avg = None
         for idx, batch in enumerate(dataloader):
+            if avg is None:
+                avg = AvgMeter()
             self.params.global_step += 1
             self.params.idx = idx
             meter = self.train_step(idx, batch, self.params)
@@ -781,15 +802,15 @@ class Trainer(DLLoopMix, _BaseTrainer):
             if self.train_epoch_toggle:
                 self.train_epoch_toggle = False
                 break
-        return avg
+        return TrainerResult(TrainerStage.train_epoch, avg, TrainerResult.MSG_OK)
 
     def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
         pass
 
-    def evaluate(self, dataloader: Union[DataLoader, DataModule] = None):
+    def evaluate(self, dataloader: Union[DataLoader, DataModule] = None) -> TrainerResult:
         dataloader = self.prepare_dataloader(TrainerStage.val, dataloader)
         if dataloader is None:
-            return TrainerResult(TrainerStage.val, 1, 'no eval_dataloader')
+            return TrainerResult(TrainerStage.val, None, TrainerResult.MSG_NO_DATALOADER)
         self._check_models_init()
         self.to_stage(TrainerStage.val)
 
@@ -799,15 +820,16 @@ class Trainer(DLLoopMix, _BaseTrainer):
                 self.params.idx = idx
                 meter = self.evaluate_step(idx, batch, self.params)
                 avg.update(self._wrap_result(meter))
-        return TrainerResult(TrainerStage.val, 0)
 
-    def evaluate_step(self, idx, batch, params: ParamsType, *args, **kwargs):
-        return self.test_step(idx, batch, params, *args, **kwargs)
+        return TrainerResult(TrainerStage.val, avg, TrainerResult.MSG_OK)
 
-    def test(self, dataloader: Union[DataLoader, DataModule] = None):
+    def evaluate_step(self, idx, batch, params: ParamsType, *args, **kwargs) -> Meter:
+        pass
+
+    def test(self, dataloader: Union[DataLoader, DataModule] = None) -> TrainerResult:
         dataloader = self.prepare_dataloader(TrainerStage.test, dataloader)
         if dataloader is None:
-            return TrainerResult(TrainerStage.test, 1, 'no test_dataloader')
+            return TrainerResult(TrainerStage.test, None, TrainerResult.MSG_NO_DATALOADER)
         self._check_models_init()
         self.to_stage(TrainerStage.test)
 
@@ -817,7 +839,7 @@ class Trainer(DLLoopMix, _BaseTrainer):
                 self.params.idx = idx
                 meter = self.test_step(idx, batch, self.params)
                 avg.update(self._wrap_result(meter))
-        return TrainerResult(TrainerStage.test, 0)
+        return TrainerResult(TrainerStage.test, avg, TrainerResult.MSG_OK)
 
     def test_step(self, idx, batch, params: ParamsType, *args, **kwargs):
         pass
