@@ -1,6 +1,4 @@
 import os
-import accelerate
-# accelerate.state.
 from torch import distributed as dist
 
 
